
# This section utilizes the RIFE (Real-Time Intermediate Flow Estimation) method
# described in:
# Huang, Z., Zhang, T., Heng, W., Shi, B., & Zhou, S. (2022). Real-time intermediate 
# flow estimation for video frame interpolation. In Proceedings of the European 
# Conference on Computer Vision (ECCV).
# The original code can be found at [https://github.com/megvii-research/ECCV2022-RIFE